﻿//
// NTFS - Myriadbits .NET NTFS library. 
// Fast MFT reader.
// Copyright (C) 2014 Myriadbits, Jochem Bakker
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// For more information, contact me at: info@myriadbits.com
//

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Myriadbits.NTFS
{
	/// <summary>
	/// This class represents a file record inside the MFT 
	/// (File record is 1024 bytes)
	/// </summary>
	public class NTFSFileRecord
	{
		private int _offset = 0;
		private DataBuffer _buffer = null;
		private List<NTFSAttribute> _attributes = null;
		private List<NTFSAttributeFilename> _nameAttributes = null;
		private long _fileSize = -1;

		public static int counter = 0;

		public NTFSFileRecord(DataBuffer buffer, int clusterSize, int offset = 0)
		{
			this._buffer = buffer;
			this._offset = offset;

			if (this.IsValid)
				ApplyFixup(clusterSize);
		}

		public bool IsValid { get { return (this._buffer.GetByte(this._offset) == 0x46 &&		// F
											this._buffer.GetByte(this._offset + 1) == 0x49 &&	// I
											this._buffer.GetByte(this._offset + 2) == 0x4C &&	// L
											this._buffer.GetByte(this._offset + 3) == 0x45);}}	// E

		/// <summary>
		/// Return a list of all attributes
		/// </summary>
		public List<NTFSAttribute> Attributes 
		{ 
			get
			{
				if (this._attributes == null)
				{
					// Get all attributes
					NTFSAttribute attribute = null;
					this._attributes = new List<NTFSAttribute>();
					int attributeOffset = this._offset + this._buffer.GetShort(this._offset + 0x14);
					do
					{
						attribute = NTFSAttributeFactory.CreateAttribute(this._buffer, attributeOffset);
						if (attribute != null)
						{
							this._attributes.Add(attribute);
							if (attribute.Length > 1024 || attribute.Length < 0) break; // Arg, error
							attributeOffset += attribute.Length;
						}
					}
					while (attribute != null);
				}
				return this._attributes;
			}
		}

		public ushort UpdateSequenceArrayOffset { get { return this._buffer.GetUShort(this._offset + 0x04); } }
		public ushort UpdateSequenceSize { get { return this._buffer.GetUShort(this._offset + 0x06); } }
		public ulong LogSequenceNumber { get { return this._buffer.GetULong(this._offset + 0x08); } }

		public ushort SequenceNumber { get { return this._buffer.GetUShort(this._offset + 0x10); } }
		public ushort HardLinkCount { get { return this._buffer.GetUShort(this._offset + 0x12); } }
		public uint Flags { get { return this._buffer.GetUShort(this._offset + 0x16); } }
		public int RealSizeOfRecord { get { return this._buffer.GetInt(this._offset + 0x18); } }
		public int AllocatedSizeOfRecord { get { return this._buffer.GetInt(this._offset + 0x1C); } }
		public ulong BaseFileRecordReference { get { return this._buffer.GetULong(this._offset + 0x20) & NTFS.referenceMask; } }

		public ulong ReferenceNumber { get { return (ulong) this._buffer.GetUInt(this._offset + 0x2C); } }
		public ulong CreateUniqueReferenceNumber(byte index0 = 0, byte index1 = 0) { return ((ulong)this._buffer.GetUInt(this._offset + 0x2C)) + ((ulong) index0 << 56) + ((ulong) index1 << 48); }

		public bool IsDirectory { get { return ((this.Flags & 0x0002) != 0x0000); } }
		public bool IsInUse { get { return ((this.Flags & 0x0001) != 0x0000); } }

		/// <summary>
		/// Use the update sequence array to fill in the original bytes
		/// </summary>
		public void ApplyFixup(int clusterSize)
		{
			// First element in update sequence array is the actual sequence number
			int usa_offset = this._offset + this.UpdateSequenceArrayOffset;
			ushort usn = this._buffer.GetUShort(usa_offset);
			if (usn != 0)
			{
				int lastBytesOfCluster = this._offset + clusterSize - 2;
				for (int n = 1; n < this.UpdateSequenceSize; n++)
				{
					ushort originalData = this._buffer.GetUShort(usa_offset + n * 2);
					this._buffer.SetUShort(lastBytesOfCluster, originalData);
					lastBytesOfCluster += clusterSize;
				}
			}
		}


		/// <summary>
		/// Return the long(est) filename
		/// </summary>
		public string Filename 
		{ 
			get
			{
				NTFSAttributeFilename attribute = this.NameAttribute;
				if (attribute != null)
					return attribute.Filename;
				return "";
			}
		}

		/// <summary>
		/// Returns the filesize of the file TODO Store in member after collecting for the first time
		/// </summary>
		/// <param name="volume"></param>
		/// <returns></returns>
		public string GetFilename(int nameIndex = 0)
		{
			// Get all name attributes with the same parent folder as the selected one
			if (nameIndex < 0 || nameIndex >= this.NameAttributes.Count)
				return "<unkown>";
			NTFSAttributeFilename indexedName = this.NameAttributes[nameIndex];
			List<NTFSAttributeFilename> names = this.NameAttributes.Where(a => a.ParentReference == indexedName.ParentReference).ToList();
			// Now get the first name who's namespace is either 0 or 1 (or 3)
			if (names.Count == 1)
			{
				// Just get the overall best filename
				foreach (NTFSAttributeFilename name in this.NameAttributes)
				{
					if (name.Namespace == 0 || name.Namespace == 1 || name.Namespace == 3)
						return name.Filename;
				}
			}
			else
			{
				// Multiple names, just get the current one
				return indexedName.Filename;
			}
			return "<unknown>";
		}
		


		/// <summary>
		/// Returns the filesize of the file TODO Store in member after collecting for the first time
		/// </summary>
		/// <param name="volume"></param>
		/// <returns></returns>
		public ulong GetFilesize(NTFSVolume volume)
		{
			if (this._fileSize == -1)
			{
				ulong dataSize = 0;

				// Start by using the data length
				if (this.HasDirectDataAttribute)
				{
					if (this.DataAttribute.NameLength == 0)
						dataSize = this.DataAttribute.DataSize; // Unnamed data = real data, use this length
				}
					
				// No datasize, try the datasiez of a names attribute
				if (dataSize == 0)
				{
					NTFSAttributeFilename name = this.NameAttributes.Where(a => a.DataSize != 0).FirstOrDefault();
					if (name != null)
						dataSize = name.DataSize;
				}

				// If still no datasize, try the data attribute by reference
				if (dataSize == 0)
				{
					if (this.DataAttributeReferenceNumber != 0)
					{
						NTFSFileRecord dataFileRecord = volume.ReadFileRecord(this.DataAttributeReferenceNumber);
						if (dataFileRecord != null && dataFileRecord.HasDirectDataAttribute)
						{
							if (dataFileRecord.DataAttribute.NameLength == 0) // (only for unnamed)
								dataSize = dataFileRecord.DataAttribute.DataSize;
						}
					}
				}

				// If still no datasize, try the names in the references
				if (dataSize == 0)
				{
					if (this.NameAttributeReferences != null)
					{
						foreach (NTFSAttributeReference nameRef in this.NameAttributeReferences)
						{
							NTFSFileRecord filenameRecord = volume.ReadFileRecord(nameRef.BaseFileReference); // Only interested in the file reference number
							if (filenameRecord != null)
							{
								NTFSAttributeFilename nameAttrRef = filenameRecord.NameAttribute;
								if (nameAttrRef != null && nameRef.StartingVCN == 0)
								{
									dataSize = nameAttrRef.RealFileSize;
									break;
								}
							}
						}
					}
				}

				this._fileSize = (long)dataSize;
			}
			return (ulong) this._fileSize;
		}


		/// <summary>
		/// Return the nameattribute with the long(est) filename
		/// </summary>
		public NTFSAttributeFilename NameAttribute
		{
			get
			{
				return NameAttributes.FirstOrDefault();
			}
		}


		/// <summary>
		/// Return the nameattribute with the long(est) filename
		/// </summary>
		public List<NTFSAttributeFilename> NameAttributes
		{
			get
			{
				if (this._nameAttributes != null)
					return this._nameAttributes;

				// Same folder reference but different namespace, only add 1
				// Same folder reference but same namespace, add each one
				this._nameAttributes = new List<NTFSAttributeFilename>();
				NTFSAttributeFilename[] fileNameAttributes = this.Attributes.Where(a => a.AttributeType == NTFSAttributeType.File_Name).Cast<NTFSAttributeFilename>().ToArray();
				
				for (int n = 0; n < fileNameAttributes.Length; n++)
				{
					NTFSAttributeFilename current = fileNameAttributes[n];
					if (current != null)
					{
						int lowestNamespace = current.Namespace;
						NTFSAttributeFilename lowest = current;
						for (int i = n + 1; i < fileNameAttributes.Length; i++)
						{
							if (fileNameAttributes[i].ParentReference == current.ParentReference)
							{
								if (fileNameAttributes[i].Namespace == lowestNamespace)
								{
									// The same namespace probably symbolic link
								}
								else if (fileNameAttributes[i].Namespace < lowestNamespace)
								{
									lowestNamespace = fileNameAttributes[i].Namespace;
									lowest = fileNameAttributes[i];
									fileNameAttributes[i] = null;
								}
								else
								{
									fileNameAttributes[i] = null;
								}
								
							}
						}
						this._nameAttributes.Add(lowest);
					}
				}
				return this._nameAttributes;				
			}
		}


		/// <summary>
		/// Return the standard attribute
		/// </summary>
		public NTFSAttributeStandardInformation StandardAttributeInformation
		{
			get
			{
				return (NTFSAttributeStandardInformation)this.Attributes.Where(a => a.AttributeType == NTFSAttributeType.Standard_Information).FirstOrDefault();
			}
		}


		/// <summary>
		/// Return the standard information attribute. When not present in this
		/// filerecord, read the base record and return that standard information
		/// </summary>
		/// <param name="volume"></param>
		/// <returns></returns>
		public NTFSAttributeStandardInformation GetStandardInformation(NTFSVolume volume)
		{
			NTFSAttributeStandardInformation si = this.StandardAttributeInformation;
			if (si == null && this.BaseFileRecordReference != 0)
			{
				NTFSFileRecord filerecordBase = volume.ReadFileRecord(this.BaseFileRecordReference);
				if (filerecordBase != null)
					si = filerecordBase.StandardAttributeInformation;
			}
			return si;
		}

		/// <summary>
		/// Return the index root (if any)
		/// </summary>
		public NTFSAttributeIndexRoot IndexRoot
		{
			get
			{
				return (NTFSAttributeIndexRoot)this.Attributes.Where(a => a.AttributeType == NTFSAttributeType.Index_Root).FirstOrDefault();
			}
		}

		/// <summary>
		/// Return the data attribute
		/// </summary>
		public bool HasDirectDataAttribute
		{
			get 
			{
				// First check if the data attribute is present in the attribute list:
				return this.Attributes.Where(a => a.AttributeType == NTFSAttributeType.Data).Count() != 0;
			}
		}

		/// <summary>
		/// Returns the DIRECT data attribute
		/// Use the GetDataAttribute method to get the actual data attribute, since the data attribute can be a reference!
		/// </summary>
		public NTFSAttributeData DataAttribute
		{
			get 
			{
				// First check if the data attribute is present in the attribute list:
				NTFSAttributeData dataAttribute = (NTFSAttributeData)this.Attributes.Where(a => a.AttributeType == NTFSAttributeType.Data).FirstOrDefault();
				return dataAttribute;
			}
		}
		
		/// <summary>
		/// Return the reference number of data attribute (when the data attribute is not directly included in this record)
		/// </summary>
		public ulong DataAttributeReferenceNumber
		{
			get 
			{
				// Not found here, check in the attributeReferences (TODO, is it possible to have 2 attribute lists?)
				NTFSAttributeAttributeList attributeList = (NTFSAttributeAttributeList)this.Attributes.Where(a => a.AttributeType == NTFSAttributeType.Attribute_List).FirstOrDefault();
				if (attributeList != null)
				{
					// Found, check for data attribute reference
					NTFSAttributeReference attributeRef = (NTFSAttributeReference)attributeList.AttributeReferences.Where(a => a.AttributeType == NTFSAttributeType.Data).FirstOrDefault();
					if (attributeRef != null)
					{
						// Tricky part, jump in  the file, hopefully this won't happen a lot...
						return attributeRef.BaseFileReference;
					}
				}
				return 0;
			}
		}


		/// <summary>
		/// Returns the Filename references attributes
		/// Use the GetDataAttribute method to get the actual data attribute, since the data attribute can be a reference!
		/// </summary>
		public List<NTFSAttributeReference> NameAttributeReferences
		{
			get
			{
				// First check if the data attribute is present in the attribute list:
				NTFSAttributeAttributeList listAttribute = (NTFSAttributeAttributeList)this.Attributes.Where(a => a.AttributeType == NTFSAttributeType.Attribute_List).FirstOrDefault();
				if (listAttribute != null)
					return listAttribute.AttributeFilenameReferences;
				return null;
			}
		}

		/// <summary>
		/// Return the data attribute
		/// </summary>
		public NTFSAttributeData GetDataAttribute(NTFSVolume volume)
		{
			// Since we want to determine the total folder size, we need the data attribute
			// The data attribute might be a reference attribute, that involves a seperate seek to read the seperate file record
			// that holds the actual data attribute.
			// On my own HD this will happen < 0.04 % of the files (116 times in 235000 files, especially large files)
			if (this.HasDirectDataAttribute)
				return this.DataAttribute;
			else
			{
				// Not found directly, get it by reference
				ulong dataReferenceNumber = this.DataAttributeReferenceNumber;
				NTFSFileRecord dataFileRecord = volume.ReadFileRecord(dataReferenceNumber);
				if (dataFileRecord != null && dataFileRecord.HasDirectDataAttribute)
					return dataFileRecord.DataAttribute;
			}
			return null;
		}


		/// <summary>
		/// Return a string representation of this object
		/// </summary>
		/// <returns></returns>
		public override string ToString()
		{
			StringBuilder sb = new StringBuilder();
			for (int n = 0; n < 0x40; n++)
				sb.Append(this._buffer.GetByte(this._offset + n).ToString("x") + ", ");
			return sb.ToString();
		}


		/// <summary>
		/// Return a string representation of this object
		/// </summary>
		/// <returns></returns>
		public StringBuilder GetInformation(NTFSVolume volume, NTFSReader reader, int level = 0)
		{
			StringBuilder sb = new StringBuilder();

			string prepend = new string(' ', level * 4);

			sb.AppendLine(string.Format("{0}File[{1}]: {2}", prepend, this.ReferenceNumber, this.Filename));

			int index = 0;
			foreach (NTFSAttributeFilename nameAttr in this.NameAttributes)
			{
				sb.AppendLine(string.Format("{0} - NameAttr[{1}]: {2}, Size {3}, Parent[{4}]: {5}", prepend, index, nameAttr.Filename, nameAttr.DataSize, nameAttr.ParentReference, reader.GetFolderPath(nameAttr.ParentReference) ));
				index++;
			}

			if (this.NameAttributeReferences != null)
			{
				index = 0;
				foreach (NTFSAttributeReference nameRef in this.NameAttributeReferences)
				{
					sb.AppendLine(string.Format("{0} - NameAttrRef[{1}]: {2}", prepend, index, nameRef.BaseFileReference));
					if (nameRef.BaseFileReference != this.ReferenceNumber)
					{
						NTFSFileRecord filenameRecord = volume.ReadFileRecord(nameRef.BaseFileReference); // Only interested in the file reference number
						if (filenameRecord != null)
						{
							sb.Append(filenameRecord.GetInformation(volume, reader, level + 1));
						}
					}
					index++;
				}
			}
			
			return sb;
		}
	}
}
